
import logging
import os
import sys
import time
import resource
import re
import multiprocessing

import graphsignal
from graphsignal.proto import profiles_pb2

logger = logging.getLogger('graphsignal')

OS_LINUX = (sys.platform.startswith('linux'))
OS_DARWIN = (sys.platform == 'darwin')
OS_WIN = (sys.platform == 'win32')
VM_RSS_REGEXP = re.compile('VmRSS:\\s+(\\d+)\\s+kB')
VM_SIZE_REGEXP = re.compile('VmSize:\\s+(\\d+)\\s+kB')


class HostReader():
    __slots__ = [
        '_last_read_time',
        '_last_cpu_time_ns'
    ]

    MIN_CPU_READ_INTERVAL = 1e9

    def __init__(self):
        self._last_read_time = None
        self._last_cpu_time_ns = None

    def setup(self):
        pass

    def shutdown(self):
        pass

    def read(self, resource_usage):
        host_usage = resource_usage.host_usage

        if not OS_WIN:
            cpu_time_ns = _read_cpu_time()
            if cpu_time_ns is not None:
                if self._last_cpu_time_ns is not None:
                    interval_ns = (time.time() - self._last_read_time) * 1e9
                    if interval_ns > HostReader.MIN_CPU_READ_INTERVAL:
                        cpu_diff_ns = cpu_time_ns - self._last_cpu_time_ns
                        cpu_usage = (cpu_diff_ns / interval_ns) * 100
                        try:
                            cpu_usage = cpu_usage / multiprocessing.cpu_count()
                        except Exception:
                            pass
                        host_usage.cpu_usage_percent = cpu_usage
                else:
                    self._last_read_time = time.time()
                    self._last_cpu_time_ns = cpu_time_ns

        if not OS_WIN:
            max_rss = _read_max_rss()
            if max_rss is not None:
                host_usage.max_rss = max_rss

        if OS_LINUX:
            current_rss = _read_current_rss()
            if current_rss is not None:
                host_usage.current_rss = current_rss

            vm_size = _read_vm_size()
            if vm_size is not None:
                host_usage.vm_size = vm_size


def _read_cpu_time():
    rusage = resource.getrusage(resource.RUSAGE_SELF)
    return int((rusage.ru_utime + rusage.ru_stime) * 1e9)  # ns


def _read_max_rss():
    rusage = resource.getrusage(resource.RUSAGE_SELF)

    if OS_DARWIN:
        return rusage.ru_maxrss
    else:
        return rusage.ru_maxrss * 1024


def _read_current_rss():
    pid = os.getpid()

    output = None
    try:
        f = open('/proc/{0}/status'.format(os.getpid()))
        output = f.read()
        f.close()
    except Exception:
        return None

    match = VM_RSS_REGEXP.search(output)
    if match:
        return int(float(match.group(1)) * 1e3)

    return None


def _read_vm_size():
    pid = os.getpid()

    output = None
    try:
        f = open('/proc/{0}/status'.format(os.getpid()))
        output = f.read()
        f.close()
    except Exception:
        return None

    match = VM_SIZE_REGEXP.search(output)
    if match:
        return int(float(match.group(1)) * 1e3)

    return None



import logging
import os
import sys
import platform
import socket
import re

from graphsignal.proto import profiles_pb2

logger = logging.getLogger('graphsignal')

version_regexp = re.compile('^(\\d+)\\.?(\\d+)?\\.?(\\d+)?')


def _read_run_env():
    run_env = profiles_pb2.RunEnvironment()

    try:
        try:
            run_env.hostname = socket.gethostname()
        except BaseException:
            logger.debug('Error reading hostname', exc_info=True)
        run_env.platform = sys.platform
        run_env.machine = platform.machine()
        run_env.os_name = os.uname().sysname
        run_env.os_version = os.uname().release
        run_env.runtime = profiles_pb2.RunEnvironment.Runtime.PYTHON
        run_env.runtime_version.major = sys.version_info.major
        run_env.runtime_version.minor = sys.version_info.minor
        run_env.runtime_version.patch = sys.version_info.micro
        run_env.runtime_impl = platform.python_implementation()
    except BaseException:
        logger.error(
            'Error reading run environment information',
            exc_info=True)

    return run_env


cached_run_env = _read_run_env()


def parse_semver(semver_proto, version):
    version_match = version_regexp.match(str(version))
    if version_match is not None:
        groups = version_match.groups()
        if groups[0] is not None:
            semver_proto.major = int(groups[0])
        if groups[1] is not None:
            semver_proto.minor = int(groups[1])
        if groups[2] is not None:
            semver_proto.patch = int(groups[2])


def compare_semver(semver_proto, version):
    semver_int = semver_proto.major * 1e6 + \
        semver_proto.minor * 1e3 + semver_proto.patch
    version_int = version[0] * 1e6 + version[1] * 1e3 + version[2]
    if semver_int < version_int:
        return -1
    if semver_int > version_int:
        return 1
    else:
        return 0



import logging

try:
    from pytorch_lightning.profiler import BaseProfiler as Profiler
except ImportError:
    from pytorch_lightning.profiler import Profiler as Profiler

import graphsignal
from graphsignal.profilers.pytorch import profile_span

logger = logging.getLogger('graphsignal')


class GraphsignalProfiler(Profiler):
    __slots__ = [
        '_span'
    ]

    def __init__(self):
        self._span = None
        super().__init__()

    def _start_profiler(self, span_name):
        if not self._span:
            self._span = profile_span(span_name=span_name)

    def _stop_profiler(self):
        if self._span:
            self._span.stop()
            self._span = None

    def start(self, action_name: str) -> None:
        if action_name == 'training_step':
            self._start_profiler('Training step')
        elif action_name == 'validation_step':
            self._start_profiler('Validation step')
        elif action_name == 'test_step':
            self._start_profiler('Test step')
        elif action_name == 'predict_step':
            self._start_profiler('Prediction step')

    def stop(self, action_name: str) -> None:
        self._stop_profiler()

    def summary(self):
        pass

    def teardown(self, stage):
        pass


import logging
import time
from threading import Lock
import traceback

import graphsignal
from graphsignal import system_info
from graphsignal.proto import profiles_pb2

logger = logging.getLogger('graphsignal')


class ProfilingSpan(object):
    __slots__ = [
        '_scheduler',
        '_profiler',
        '_is_scheduled',
        '_is_profiling',
        '_profile',
        '_metadata',
        '_stop_lock'
    ]

    SPAN_TIMEOUT_SEC = 10

    def __init__(self, scheduler, profiler, span_name=None, ensure_profile=False):
        self._scheduler = scheduler
        self._profiler = profiler
        self._is_scheduled = False
        self._is_profiling = False
        self._profile = None
        self._metadata = None
        self._stop_lock = Lock()

        if self._scheduler.lock(ensure=ensure_profile):
            self._is_scheduled = True
            self._profile = profiles_pb2.MLProfile()
            self._profile.workload_name = graphsignal._agent.workload_name
            self._profile.run_id = graphsignal._agent.run_id
            self._profile.run_start_ms = graphsignal._agent.run_start_ms
            if span_name:
                self._profile.span_name = span_name

            try:
                self._profile.run_env.CopyFrom(system_info.cached_run_env)
                self._profile.resource_usage.read_us = _timestamp_us()
                graphsignal._agent.host_reader.read(self._profile.resource_usage)
                graphsignal._agent.nvml_reader.read(self._profile.resource_usage)
            except Exception as exc:
                _add_exception(self._profile, exc)
                logger.error(
                    'Error taking run snapshot', exc_info=True)

            try:
                if self._profiler.start():
                    self._is_profiling = True
                else:
                    self._scheduler.unlock()
            except Exception as exc:
                self._scheduler.unlock()
                self._is_profiling = False
                _add_exception(self._profile, exc)
                logger.error('Error starting profiler', exc_info=True)

            self._profile.start_us = _timestamp_us()

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.stop()

    def set_name(self, name):
        if self._is_scheduled:
            if name:
                self._profile.span_name = name

    def add_metadata(self, key, value):
        if self._is_scheduled:
            if self._metadata is None:
                self._metadata = {}
            self._metadata[key] = value

    def stop(self, no_save=False):
        with self._stop_lock:
            if self._is_scheduled:
                self._profile.end_us = _timestamp_us()
                if self._metadata is not None:
                    for key, value in self._metadata.items():
                        entry = self._profile.metadata.add()
                        entry.key = str(key)
                        entry.value = str(value)

                try:
                    if self._is_profiling:
                        if self._profiler.stop(self._profile):
                            if not no_save:
                                _upload_profile(self._profile)
                    else:
                        if not no_save:
                            _upload_profile(self._profile)
                except Exception as exc:
                    logger.error('Error stopping profiler', exc_info=True)
                    _add_exception(self._profile, exc)
                    _upload_profile(self._profile)
                finally:
                    self._is_scheduled = False
                    self._is_profiling = False
                    self._profile = None
                    self._scheduler.unlock()


def _add_exception(profile, exc):
    profiler_error = profile.profiler_errors.add()
    profiler_error.message = str(exc)
    if exc.__traceback__:
        frames = traceback.format_tb(exc.__traceback__)
        if len(frames) > 0:
            profiler_error.stack_trace = ''.join(frames)


def _upload_profile(profile):
    graphsignal._agent.uploader.upload_profile(profile)
    graphsignal._agent.uploader.flush_in_thread()


def _timestamp_us():
    return int(time.time() * 1e6)




class Agent(object):
    def __init__(self):
        self.run_id = None
        self.run_start_ms = None
        self.api_key = None
        self.workload_name = None
        self.debug_mode = None
        self.uploader = None
        self.span_scheduler = None
        self.profiler = None
        self.active_span = None
        self.host_reader = None
        self.nvml_reader = None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: profiles.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0eprofiles.proto\x12\x14graphsignal.profiles\"X\n\rUploadRequest\x12\x34\n\x0bml_profiles\x18\x01 \x03(\x0b\x32\x1f.graphsignal.profiles.MLProfile\x12\x11\n\tupload_ms\x18\x02 \x01(\x04\"\x10\n\x0eUploadResponse\"\x9a\x01\n\rProfileRecord\x12\x0f\n\x07\x64\x61ta_id\x18\x01 \x01(\t\x12\x13\n\x0bworkload_id\x18\x02 \x01(\t\x12\x12\n\nprofile_id\x18\x03 \x01(\t\x12\x33\n\nml_profile\x18\x04 \x01(\x0b\x32\x1f.graphsignal.profiles.MLProfile\x12\x1a\n\x12\x64\x61ta_retention_sec\x18\x05 \x01(\x04\"\x86\x04\n\tMLProfile\x12\x15\n\rworkload_name\x18\x01 \x01(\t\x12\x0e\n\x06run_id\x18\x02 \x01(\t\x12\x14\n\x0crun_start_ms\x18\x03 \x01(\x04\x12\x11\n\tspan_name\x18\x04 \x01(\t\x12\x10\n\x08start_us\x18\x05 \x01(\x04\x12\x0e\n\x06\x65nd_us\x18\x06 \x01(\x04\x12<\n\x0fprofiler_errors\x18\x07 \x03(\x0b\x32#.graphsignal.profiles.ProfilerError\x12\x30\n\x08metadata\x18\x08 \x03(\x0b\x32\x1e.graphsignal.profiles.Metadata\x12\x35\n\x07run_env\x18\t \x01(\x0b\x32$.graphsignal.profiles.RunEnvironment\x12;\n\x0eresource_usage\x18\r \x01(\x0b\x32#.graphsignal.profiles.ResourceUsage\x12\x39\n\x07summary\x18\n \x01(\x0b\x32(.graphsignal.profiles.PerformanceSummary\x12/\n\x08op_stats\x18\x0b \x03(\x0b\x32\x1d.graphsignal.profiles.OpStats\x12\x37\n\x0ckernel_stats\x18\x0c \x03(\x0b\x32!.graphsignal.profiles.KernelStats\"&\n\x08Metadata\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\x9f\x04\n\x0eRunEnvironment\x12\x10\n\x08hostname\x18\x01 \x01(\t\x12\x10\n\x08platform\x18\x02 \x01(\t\x12\x0f\n\x07machine\x18\x03 \x01(\t\x12\x0f\n\x07os_name\x18\x04 \x01(\t\x12\x12\n\nos_version\x18\x05 \x01(\t\x12=\n\x07runtime\x18\x06 \x01(\x0e\x32,.graphsignal.profiles.RunEnvironment.Runtime\x12\x35\n\x0fruntime_version\x18\x07 \x01(\x0b\x32\x1c.graphsignal.profiles.SemVer\x12\x14\n\x0cruntime_impl\x18\x08 \x01(\t\x12\x46\n\x0cml_framework\x18\t \x01(\x0e\x32\x30.graphsignal.profiles.RunEnvironment.MLFramework\x12:\n\x14ml_framework_version\x18\n \x01(\x0b\x32\x1c.graphsignal.profiles.SemVer\x12-\n\x07\x64\x65vices\x18\x0b \x03(\x0b\x32\x1c.graphsignal.profiles.Device\",\n\x07Runtime\x12\x15\n\x11RUNTIME_UNDEFINED\x10\x00\x12\n\n\x06PYTHON\x10\x01\"F\n\x0bMLFramework\x12\x1a\n\x16ML_FRAMEWORK_UNDEFINED\x10\x00\x12\x0e\n\nTENSORFLOW\x10\x01\x12\x0b\n\x07PYTORCH\x10\x02\"5\n\x06SemVer\x12\r\n\x05major\x18\x01 \x01(\x05\x12\r\n\x05minor\x18\x02 \x01(\x05\x12\r\n\x05patch\x18\x03 \x01(\x05\"\xaf\x01\n\x06\x44\x65vice\x12.\n\x04type\x18\x01 \x01(\x0e\x32 .graphsignal.profiles.DeviceType\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x17\n\x0fis_cuda_enabled\x18\x03 \x01(\x08\x12\x38\n\x12\x63ompute_capability\x18\x04 \x01(\x0b\x32\x1c.graphsignal.profiles.SemVer\x12\x14\n\x0ctotal_memory\x18\x05 \x01(\x04\"\x8e\x01\n\rResourceUsage\x12\x0f\n\x07read_us\x18\x01 \x01(\x04\x12\x33\n\nhost_usage\x18\x02 \x01(\x0b\x32\x1f.graphsignal.profiles.HostUsage\x12\x37\n\x0c\x64\x65vice_usage\x18\x03 \x03(\x0b\x32!.graphsignal.profiles.DeviceUsage\"]\n\tHostUsage\x12\x19\n\x11\x63pu_usage_percent\x18\x01 \x01(\x01\x12\x0f\n\x07max_rss\x18\x02 \x01(\x04\x12\x13\n\x0b\x63urrent_rss\x18\x03 \x01(\x04\x12\x0f\n\x07vm_size\x18\x04 \x01(\x04\"\xf4\x01\n\x0b\x44\x65viceUsage\x12\x11\n\tdevice_id\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65vice_name\x18\x02 \x01(\t\x12\x11\n\tmem_total\x18\x03 \x01(\x04\x12\x10\n\x08mem_used\x18\x04 \x01(\x04\x12\x10\n\x08mem_free\x18\x05 \x01(\x04\x12\x1f\n\x17gpu_utilization_percent\x18\x06 \x01(\x01\x12\x1f\n\x17mem_utilization_percent\x18\x07 \x01(\x01\x12\x12\n\ngpu_temp_c\x18\x08 \x01(\x01\x12\x15\n\rpower_usage_w\x18\t \x01(\x01\x12\x19\n\x11\x66\x61n_speed_percent\x18\n \x01(\x01\"\x94\x02\n\x12PerformanceSummary\x12\x1b\n\x13\x64\x65vice_idle_percent\x18\x01 \x01(\x01\x12\x19\n\x11host_idle_percent\x18\x02 \x01(\x01\x12$\n\x1c\x64\x65vice_compute_16bit_percent\x18\x03 \x01(\x01\x12$\n\x1c\x64\x65vice_compute_32bit_percent\x18\x04 \x01(\x01\x12\x17\n\x0fhost_op_percent\x18\x06 \x01(\x01\x12\x19\n\x11\x64\x65vice_op_percent\x18\x07 \x01(\x01\x12\x17\n\x0fmxu_utilization\x18\x08 \x01(\x01\x12\x15\n\rinput_percent\x18\t \x01(\x01\x12\x16\n\x0eoutput_percent\x18\n \x01(\x01\"\xd3\x03\n\x07OpStats\x12\x35\n\x0b\x64\x65vice_type\x18\x01 \x01(\x0e\x32 .graphsignal.profiles.DeviceType\x12\x11\n\tdevice_id\x18\x02 \x01(\t\x12\x0f\n\x07op_type\x18\x03 \x01(\t\x12\x0f\n\x07op_name\x18\x04 \x01(\t\x12\r\n\x05\x63ount\x18\x05 \x01(\x04\x12\x1a\n\x12total_host_time_us\x18\x06 \x01(\x04\x12\x1c\n\x14total_device_time_us\x18\x07 \x01(\x04\x12\x19\n\x11self_host_time_us\x18\x08 \x01(\x04\x12\x1b\n\x13self_device_time_us\x18\t \x01(\x04\x12\x19\n\x11total_host_memory\x18\n \x01(\x04\x12\x1b\n\x13total_device_memory\x18\x0b \x01(\x04\x12\x18\n\x10self_host_memory\x18\x0c \x01(\x04\x12\x1a\n\x12self_device_memory\x18\r \x01(\x04\x12\x1d\n\x15self_host_memory_rate\x18\x0e \x01(\x04\x12\x1f\n\x17self_device_memory_rate\x18\x0f \x01(\x04\x12\r\n\x05\x66lops\x18\x10 \x01(\x04\x12\x1e\n\x16tensorcore_utilization\x18\x11 \x01(\x01\"\xbe\x01\n\x0bKernelStats\x12\x35\n\x0b\x64\x65vice_type\x18\x01 \x01(\x0e\x32 .graphsignal.profiles.DeviceType\x12\x11\n\tdevice_id\x18\x02 \x01(\t\x12\x0f\n\x07op_name\x18\x03 \x01(\t\x12\x13\n\x0bkernel_name\x18\x04 \x01(\t\x12\r\n\x05\x63ount\x18\x05 \x01(\x04\x12\x13\n\x0b\x64uration_ns\x18\x06 \x01(\x04\x12\x1b\n\x13is_using_tensorcore\x18\x07 \x01(\x08\"5\n\rProfilerError\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x13\n\x0bstack_trace\x18\x02 \x01(\t*B\n\nDeviceType\x12\x19\n\x15\x44\x45VICE_TYPE_UNDEFINED\x10\x00\x12\x07\n\x03\x43PU\x10\x01\x12\x07\n\x03GPU\x10\x02\x12\x07\n\x03TPU\x10\x03\x62\x06proto3')

_DEVICETYPE = DESCRIPTOR.enum_types_by_name['DeviceType']
DeviceType = enum_type_wrapper.EnumTypeWrapper(_DEVICETYPE)
DEVICE_TYPE_UNDEFINED = 0
CPU = 1
GPU = 2
TPU = 3


_UPLOADREQUEST = DESCRIPTOR.message_types_by_name['UploadRequest']
_UPLOADRESPONSE = DESCRIPTOR.message_types_by_name['UploadResponse']
_PROFILERECORD = DESCRIPTOR.message_types_by_name['ProfileRecord']
_MLPROFILE = DESCRIPTOR.message_types_by_name['MLProfile']
_METADATA = DESCRIPTOR.message_types_by_name['Metadata']
_RUNENVIRONMENT = DESCRIPTOR.message_types_by_name['RunEnvironment']
_SEMVER = DESCRIPTOR.message_types_by_name['SemVer']
_DEVICE = DESCRIPTOR.message_types_by_name['Device']
_RESOURCEUSAGE = DESCRIPTOR.message_types_by_name['ResourceUsage']
_HOSTUSAGE = DESCRIPTOR.message_types_by_name['HostUsage']
_DEVICEUSAGE = DESCRIPTOR.message_types_by_name['DeviceUsage']
_PERFORMANCESUMMARY = DESCRIPTOR.message_types_by_name['PerformanceSummary']
_OPSTATS = DESCRIPTOR.message_types_by_name['OpStats']
_KERNELSTATS = DESCRIPTOR.message_types_by_name['KernelStats']
_PROFILERERROR = DESCRIPTOR.message_types_by_name['ProfilerError']
_RUNENVIRONMENT_RUNTIME = _RUNENVIRONMENT.enum_types_by_name['Runtime']
_RUNENVIRONMENT_MLFRAMEWORK = _RUNENVIRONMENT.enum_types_by_name['MLFramework']
UploadRequest = _reflection.GeneratedProtocolMessageType('UploadRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADREQUEST,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.UploadRequest)
  })
_sym_db.RegisterMessage(UploadRequest)

UploadResponse = _reflection.GeneratedProtocolMessageType('UploadResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADRESPONSE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.UploadResponse)
  })
_sym_db.RegisterMessage(UploadResponse)

ProfileRecord = _reflection.GeneratedProtocolMessageType('ProfileRecord', (_message.Message,), {
  'DESCRIPTOR' : _PROFILERECORD,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.ProfileRecord)
  })
_sym_db.RegisterMessage(ProfileRecord)

MLProfile = _reflection.GeneratedProtocolMessageType('MLProfile', (_message.Message,), {
  'DESCRIPTOR' : _MLPROFILE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.MLProfile)
  })
_sym_db.RegisterMessage(MLProfile)

Metadata = _reflection.GeneratedProtocolMessageType('Metadata', (_message.Message,), {
  'DESCRIPTOR' : _METADATA,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.Metadata)
  })
_sym_db.RegisterMessage(Metadata)

RunEnvironment = _reflection.GeneratedProtocolMessageType('RunEnvironment', (_message.Message,), {
  'DESCRIPTOR' : _RUNENVIRONMENT,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.RunEnvironment)
  })
_sym_db.RegisterMessage(RunEnvironment)

SemVer = _reflection.GeneratedProtocolMessageType('SemVer', (_message.Message,), {
  'DESCRIPTOR' : _SEMVER,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.SemVer)
  })
_sym_db.RegisterMessage(SemVer)

Device = _reflection.GeneratedProtocolMessageType('Device', (_message.Message,), {
  'DESCRIPTOR' : _DEVICE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.Device)
  })
_sym_db.RegisterMessage(Device)

ResourceUsage = _reflection.GeneratedProtocolMessageType('ResourceUsage', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEUSAGE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.ResourceUsage)
  })
_sym_db.RegisterMessage(ResourceUsage)

HostUsage = _reflection.GeneratedProtocolMessageType('HostUsage', (_message.Message,), {
  'DESCRIPTOR' : _HOSTUSAGE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.HostUsage)
  })
_sym_db.RegisterMessage(HostUsage)

DeviceUsage = _reflection.GeneratedProtocolMessageType('DeviceUsage', (_message.Message,), {
  'DESCRIPTOR' : _DEVICEUSAGE,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.DeviceUsage)
  })
_sym_db.RegisterMessage(DeviceUsage)

PerformanceSummary = _reflection.GeneratedProtocolMessageType('PerformanceSummary', (_message.Message,), {
  'DESCRIPTOR' : _PERFORMANCESUMMARY,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.PerformanceSummary)
  })
_sym_db.RegisterMessage(PerformanceSummary)

OpStats = _reflection.GeneratedProtocolMessageType('OpStats', (_message.Message,), {
  'DESCRIPTOR' : _OPSTATS,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.OpStats)
  })
_sym_db.RegisterMessage(OpStats)

KernelStats = _reflection.GeneratedProtocolMessageType('KernelStats', (_message.Message,), {
  'DESCRIPTOR' : _KERNELSTATS,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.KernelStats)
  })
_sym_db.RegisterMessage(KernelStats)

ProfilerError = _reflection.GeneratedProtocolMessageType('ProfilerError', (_message.Message,), {
  'DESCRIPTOR' : _PROFILERERROR,
  '__module__' : 'profiles_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.profiles.ProfilerError)
  })
_sym_db.RegisterMessage(ProfilerError)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _DEVICETYPE._serialized_start=3129
  _DEVICETYPE._serialized_end=3195
  _UPLOADREQUEST._serialized_start=40
  _UPLOADREQUEST._serialized_end=128
  _UPLOADRESPONSE._serialized_start=130
  _UPLOADRESPONSE._serialized_end=146
  _PROFILERECORD._serialized_start=149
  _PROFILERECORD._serialized_end=303
  _MLPROFILE._serialized_start=306
  _MLPROFILE._serialized_end=824
  _METADATA._serialized_start=826
  _METADATA._serialized_end=864
  _RUNENVIRONMENT._serialized_start=867
  _RUNENVIRONMENT._serialized_end=1410
  _RUNENVIRONMENT_RUNTIME._serialized_start=1294
  _RUNENVIRONMENT_RUNTIME._serialized_end=1338
  _RUNENVIRONMENT_MLFRAMEWORK._serialized_start=1340
  _RUNENVIRONMENT_MLFRAMEWORK._serialized_end=1410
  _SEMVER._serialized_start=1412
  _SEMVER._serialized_end=1465
  _DEVICE._serialized_start=1468
  _DEVICE._serialized_end=1643
  _RESOURCEUSAGE._serialized_start=1646
  _RESOURCEUSAGE._serialized_end=1788
  _HOSTUSAGE._serialized_start=1790
  _HOSTUSAGE._serialized_end=1883
  _DEVICEUSAGE._serialized_start=1886
  _DEVICEUSAGE._serialized_end=2130
  _PERFORMANCESUMMARY._serialized_start=2133
  _PERFORMANCESUMMARY._serialized_end=2409
  _OPSTATS._serialized_start=2412
  _OPSTATS._serialized_end=2879
  _KERNELSTATS._serialized_start=2882
  _KERNELSTATS._serialized_end=3072
  _PROFILERERROR._serialized_start=3074
  _PROFILERERROR._serialized_end=3127
# @@protoc_insertion_point(module_scope)



import unittest
import logging
import sys
from unittest.mock import patch, Mock
import tensorflow as tf
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.profilers.tensorflow import TensorflowProfiler
from graphsignal.proto import profiles_pb2

logger = logging.getLogger('graphsignal')


class TensorflowProfilerTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            workload_name='w1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_start_stop(self):
        profiler = TensorflowProfiler()

        @tf.function
        def f(x):
            while tf.reduce_sum(x) > 1:
                #tf.print(x)
                x = tf.tanh(x)
            return x

        profiler.start()
        f(tf.random.uniform([5]))
        profile = profiles_pb2.MLProfile()
        profiler.stop(profile)

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(profile))

        self.assertEqual(
            profile.run_env.ml_framework,
            profiles_pb2.RunEnvironment.MLFramework.TENSORFLOW)
        if len(tf.config.list_physical_devices('GPU')) > 0:
            self.assertEqual(
                profile.run_env.devices[0].type,
                profiles_pb2.DeviceType.GPU)
        else:
            self.assertEqual(len(profile.run_env.devices), 0)

        test_op_stats = None
        for op_stats in profile.op_stats:
            if op_stats.op_name == 'RandomUniform':
                test_op_stats = op_stats
                break
        self.assertIsNotNone(test_op_stats)
        self.assertTrue(test_op_stats.count >= 1)
        if len(tf.config.list_physical_devices('GPU')) > 0:
            self.assertTrue(profile.summary.device_idle_percent > 0)
            self.assertTrue(profile.summary.device_op_percent > 0)
            self.assertTrue(test_op_stats.total_device_time_us >= 1)
            self.assertTrue(test_op_stats.self_device_time_us >= 1)
        else:
            self.assertTrue(profile.summary.host_idle_percent > 0)
            self.assertTrue(profile.summary.host_op_percent > 0)
            self.assertTrue(test_op_stats.total_host_time_us >= 1)
            self.assertTrue(test_op_stats.self_host_time_us >= 1)



import unittest
import logging
import sys
from unittest.mock import patch, Mock

import graphsignal

logger = logging.getLogger('graphsignal')


class GraphsignalTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            workload_name='w1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_configure(self):
        self.assertEqual(graphsignal._agent.api_key, 'k1')
        self.assertEqual(graphsignal._agent.workload_name, 'w1')
        self.assertTrue(graphsignal._agent.run_start_ms > 0)
        self.assertEqual(graphsignal._agent.debug_mode, True)



import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.proto import profiles_pb2
from graphsignal.profiling_span import ProfilingSpan
from graphsignal.span_scheduler import SpanScheduler
from graphsignal.profilers.tensorflow import TensorflowProfiler
from graphsignal.usage.host_reader import HostReader
from graphsignal.usage.nvml_reader import NvmlReader
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class ProflingSpanTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            workload_name='w1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    @patch.object(TensorflowProfiler, 'start', return_value=True)
    @patch.object(TensorflowProfiler, 'stop', return_value=True)
    @patch.object(HostReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_profile')
    def test_start_stop(self, mocked_upload_profile, mocked_nvml_read, mocked_host_read,
                        mocked_stop, mocked_start):
        span = ProfilingSpan(
            scheduler=SpanScheduler(),
            profiler=TensorflowProfiler(),
            span_name='s1',
            ensure_profile=True)
        span.add_metadata('m1', 'v1')
        span.add_metadata('m2', 'v2')
        span.stop()

        mocked_start.assert_called_once()
        mocked_stop.assert_called_once()
        profile = mocked_upload_profile.call_args[0][0]

        self.assertEqual(profile.workload_name, 'w1')
        self.assertTrue(profile.run_id != '')
        self.assertEqual(profile.span_name, 's1')
        self.assertTrue(profile.start_us > 0)
        self.assertTrue(profile.end_us > 0)
        self.assertEqual(profile.metadata[0].key, 'm1')
        self.assertEqual(profile.metadata[0].value, 'v1')
        self.assertEqual(profile.metadata[1].key, 'm2')
        self.assertEqual(profile.metadata[1].value, 'v2')

    @patch.object(TensorflowProfiler, 'start', return_value=True)
    @patch.object(TensorflowProfiler, 'stop', return_value=True)
    @patch.object(HostReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_profile')
    def test_start_exception(self, mocked_upload_profile, mocked_nvml_read, mocked_host_read,
                             mocked_stop, mocked_start):
        mocked_start.side_effect = Exception('ex1')
        span = ProfilingSpan(
            scheduler=SpanScheduler(),
            profiler=TensorflowProfiler(),
            span_name='s1',
            ensure_profile=True)
        span.stop()

        mocked_start.assert_called_once()
        mocked_stop.assert_not_called()
        profile = mocked_upload_profile.call_args[0][0]

        self.assertEqual(profile.workload_name, 'w1')
        self.assertTrue(profile.run_id != '')
        self.assertEqual(profile.span_name, 's1')
        self.assertTrue(profile.start_us > 0)
        self.assertTrue(profile.end_us > 0)
        self.assertEqual(profile.profiler_errors[0].message, 'ex1')
        self.assertNotEqual(profile.profiler_errors[0].stack_trace, '')

    @patch.object(TensorflowProfiler, 'start', return_value=True)
    @patch.object(TensorflowProfiler, 'stop', return_value=True)
    @patch.object(HostReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_profile')
    def test_stop_exception(self, mocked_upload_profile, mocked_nvml_read, mocked_host_read,
                            mocked_stop, mocked_start):
        mocked_stop.side_effect = Exception('ex1')
        span = ProfilingSpan(
            scheduler=SpanScheduler(),
            profiler=TensorflowProfiler(),
            span_name='s1',
            ensure_profile=True)
        span.stop()

        mocked_start.assert_called_once()
        mocked_stop.assert_called_once()
        profile = mocked_upload_profile.call_args[0][0]

        self.assertEqual(profile.workload_name, 'w1')
        self.assertTrue(profile.run_id != '')
        self.assertEqual(profile.span_name, 's1')
        self.assertTrue(profile.start_us > 0)
        self.assertTrue(profile.end_us > 0)
        self.assertEqual(profile.profiler_errors[0].message, 'ex1')
        self.assertNotEqual(profile.profiler_errors[0].stack_trace, '')



import logging
import os
import sys
import time
import tempfile
import shutil
import glob
import torch
from torch.autograd import DeviceType

import graphsignal
from graphsignal.system_info import parse_semver
from graphsignal.proto import profiles_pb2
from graphsignal.profiling_span import ProfilingSpan

logger = logging.getLogger('graphsignal')


class PyTorchProfiler():
    __slots__ = [
        '_torch_prof',
        '_run_env'
    ]

    def __init__(self):
        self._torch_prof = None
        self._run_env = None

    def start(self):
        logger.debug('Activating PyTorch profiler')

        if not self._torch_prof:
            logger.debug('Warming up PyTorch profiler before first use')

            def _schedule_func(step):
                return torch.profiler.ProfilerAction.RECORD

            self._torch_prof = torch.profiler.profile(
                schedule=_schedule_func,
                record_shapes=False,
                with_stack=False,
                with_flops=True)
            self._torch_prof.start()
            self._torch_prof.stop()
            logger.debug('Finished warming up')

            self._read_run_env()

        self._torch_prof.start()

        return True

    def stop(self, profile):
        logger.debug('Deactivating PyTorch profiler')

        self._torch_prof.stop()

        self._copy_run_env(profile)
        self._convert_to_profile(profile)

        return True

    def _read_run_env(self):
        self._run_env = profiles_pb2.RunEnvironment()
        self._run_env.ml_framework = profiles_pb2.RunEnvironment.MLFramework.Value('PYTORCH')
        parse_semver(self._run_env.ml_framework_version, torch.__version__)
        if torch.cuda.is_available():
            for index in range(torch.cuda.device_count()):
                device_proto = self._run_env.devices.add()
                device_proto.type = profiles_pb2.DeviceType.GPU
                device_proto.name = torch.cuda.get_device_name(index)
                device_proto.is_cuda_enabled = True

                device_properties = torch.cuda.get_device_properties(index)
                device_proto.compute_capability.major = device_properties.major
                device_proto.compute_capability.minor = device_properties.minor
                device_proto.total_memory = device_properties.total_memory

    def _copy_run_env(self, profile):
        profile.run_env.ml_framework = self._run_env.ml_framework
        profile.run_env.ml_framework_version.CopyFrom(
            self._run_env.ml_framework_version)
        profile.run_env.devices.extend(self._run_env.devices)

    def _convert_to_profile(self, profile):
        for event_avg in self._torch_prof.key_averages():
            if event_avg.key and event_avg.key.startswith('ProfilerStep'):
                continue
            op_stats = profile.op_stats.add()
            if event_avg.device_type == DeviceType.CUDA:
                op_stats.device_type = profiles_pb2.DeviceType.GPU
            else:
                op_stats.device_type = profiles_pb2.DeviceType.CPU
            op_stats.op_name = event_avg.key
            op_stats.count = int(event_avg.count)
            op_stats.total_host_time_us = int(event_avg.cpu_time_total)
            op_stats.total_device_time_us = int(event_avg.cuda_time_total)
            op_stats.self_host_time_us = int(event_avg.self_cpu_time_total)
            op_stats.self_device_time_us = int(event_avg.self_cuda_time_total)
            op_stats.total_host_memory = int(event_avg.cpu_memory_usage)
            op_stats.total_device_memory = int(event_avg.cuda_memory_usage)
            op_stats.self_host_memory = int(event_avg.self_cpu_memory_usage)
            op_stats.self_device_memory = int(event_avg.self_cuda_memory_usage)
            op_stats.flops = int(event_avg.flops)

        kernel_index = {}
        for event in self._torch_prof.events():
            for kernel in event.kernels:
                key = (event.key, kernel.name, kernel.device)
                if key in kernel_index:
                    kernel_stats = kernel_index[key]
                    kernel_stats.count += 1
                    kernel_stats.duration_ns += int(kernel.duration * 1000)
                else:
                    kernel_stats = kernel_index[key] = profile.kernel_stats.add()
                    kernel_stats.device_type = profiles_pb2.DeviceType.GPU
                    kernel_stats.device_id = str(kernel.device)
                    kernel_stats.op_name = event.name
                    kernel_stats.kernel_name = kernel.name
                    kernel_stats.count = 1
                    kernel_stats.duration_ns = int(kernel.duration * 1000)

        for kernel_stats in kernel_index.values():
            profile.kernel_stats.append(kernel_stats)

        sum_host_op_time_us = 0
        sum_device_op_time_us = 0
        for op_stats in profile.op_stats:
            sum_host_op_time_us += op_stats.self_host_time_us
            sum_device_op_time_us += op_stats.self_device_time_us
        sum_op_time_us = sum_host_op_time_us + sum_device_op_time_us
        if sum_op_time_us > 0:
            profile.summary.host_op_percent = sum_host_op_time_us / sum_op_time_us * 100
            profile.summary.device_op_percent = sum_device_op_time_us / sum_op_time_us * 100

        logger.debug(
            'Converted %d PyTorch operation statistics', len(profile.op_stats))


_profiler = PyTorchProfiler()

def profile_span(span_name=None, ensure_profile=False):
    graphsignal._check_configured()

    return ProfilingSpan(
        scheduler=graphsignal._agent.span_scheduler,
        profiler=_profiler,
        span_name=span_name,
        ensure_profile=ensure_profile)



import unittest
import logging
import sys
from unittest.mock import patch, Mock
import torch
from google.protobuf.json_format import MessageToJson
import pprint
import time

import graphsignal
from graphsignal.usage.nvml_reader import NvmlReader
from graphsignal.proto import profiles_pb2

logger = logging.getLogger('graphsignal')


class NvmlReaderTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            workload_name='w1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_read(self):
        resource_usage = profiles_pb2.ResourceUsage()
        reader = graphsignal._agent.nvml_reader
        reader.read(resource_usage)

        #pp = pprint.PrettyPrinter()
        # pp.pprint(MessageToJson(resource_usage))

        if len(resource_usage.device_usage) > 0:
            device_usage = resource_usage.device_usage[0]
            self.assertNotEqual(device_usage.device_id, '')
            self.assertNotEqual(device_usage.device_name, '')
            self.assertTrue(device_usage.mem_total > 0)
            self.assertTrue(device_usage.mem_used > 0)
            self.assertTrue(device_usage.mem_free > 0)
            #self.assertTrue(device_usage.gpu_utilization_percent > 0)
            #self.assertTrue(device_usage.mem_utilization_percent > 0)
            self.assertTrue(device_usage.gpu_temp_c > 0)
            self.assertTrue(device_usage.power_usage_w > 0)
            #self.assertTrue(device_usage.fan_speed_percent > 0)



import logging
from transformers import TrainerCallback

import graphsignal.profilers.tensorflow import profile_span as profile_span_tf
import graphsignal.profilers.pytorch import profile_span as profile_span_pt

logger = logging.getLogger('graphsignal')


class GraphsignalTFCallback(TrainerCallback):
    __slots__ = [
        '_span'
    ]

    def __init__(self):
        self._span = None

    def _start_profiler(self, span_name, batch=None):
        if not self._span:
            self._span = profile_span_tf(span_name=span_name)
            if batch is not None:
                self._span.add_metadata('batch', batch)

    def _stop_profiler(self):
        if self._span:
            self._span.stop()
            self._span = None

    def on_step_begin(self, args, state, control, **kwarg):
        self._start_profiler('Training step')

    def on_step_end(self, args, state, control, **kwarg):
        self._stop_profiler()


class GraphsignalPTCallback(TrainerCallback):
    __slots__ = [
        '_span'
    ]

    def __init__(self):
        self._span = None

    def _start_profiler(self, span_name, batch=None):
        if not self._span:
            self._span = profile_span_pt(span_name=span_name)
            if batch is not None:
                self._span.add_metadata('batch', batch)

    def _stop_profiler(self):
        if self._span:
            self._span.stop()
            self._span = None

    def on_step_begin(self, args, state, control, **kwarg):
        self._start_profiler('Training step')

    def on_step_end(self, args, state, control, **kwarg):
        self._stop_profiler()



import logging
import os
import sys
import time

import graphsignal
from graphsignal.proto import profiles_pb2
from graphsignal.usage.pynvml import *

logger = logging.getLogger('graphsignal')


class NvmlReader():
    __slots__ = [
        '_is_initialized'
    ]

    def __init__(self):
        self._is_initialized = False

    def setup(self):
        if self._is_initialized:
            return
        try:
            nvmlInit()
            self._is_initialized = True
            logger.debug('Initialized NVML')
        except BaseException:
            logger.debug('Error initializing NVML, skipping GPU usage')

    def shutdown(self):
        if not self._is_initialized:
            return
        try:
            nvmlShutdown()
            self._is_initialized = False
        except BaseException:
            logger.error('Error shutting down NVML', exc_info=True)

    def read(self, resource_usage):
        if not self._is_initialized:
            return

        device_count = nvmlDeviceGetCount()
        for i in range(0, device_count):
            device_usage = resource_usage.device_usage.add()
            handle = nvmlDeviceGetHandleByIndex(i)

            try:
                pci_info = nvmlDeviceGetPciInfo(handle)
                device_usage.device_id = pci_info.busId
            except NVMLError as err:
                log_nvml_error(err)

            try:
                device_usage.device_name = nvmlDeviceGetName(handle)
            except NVMLError as err:
                log_nvml_error(err)

            try:
                mem_info = nvmlDeviceGetMemoryInfo(handle)
                device_usage.mem_total = mem_info.total
                device_usage.mem_used = mem_info.used
                device_usage.mem_free = mem_info.total - mem_info.used
            except NVMLError as err:
                log_nvml_error(err)

            try:
                util_rates = nvmlDeviceGetUtilizationRates(handle)
                device_usage.gpu_utilization_percent = util_rates.gpu
                device_usage.mem_utilization_percent = util_rates.memory
            except NVMLError as err:
                log_nvml_error(err)

            try:
                device_usage.gpu_temp_c = nvmlDeviceGetTemperature(
                    handle, NVML_TEMPERATURE_GPU)
            except NVMLError as err:
                log_nvml_error(err)

            try:
                device_usage.power_usage_w = nvmlDeviceGetPowerUsage(
                    handle) / 1000.0
            except NVMLError as err:
                log_nvml_error(err)

            try:
                device_usage.fan_speed_percent = nvmlDeviceGetFanSpeed(
                    handle)
            except NVMLError as err:
                log_nvml_error(err)


def log_nvml_error(err):
    if (err.value == NVML_ERROR_NOT_SUPPORTED):
        logger.debug('NVML call not supported')
    else:
        logger.error('Error calling NVML', exc_info=True)



import unittest
import logging
import sys
from unittest.mock import patch, Mock
import torch
from google.protobuf.json_format import MessageToJson
import pprint
import time

import graphsignal
from graphsignal.usage.host_reader import HostReader
from graphsignal.proto import profiles_pb2

logger = logging.getLogger('graphsignal')


class HostReaderTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            workload_name='w1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_read(self):
        resource_usage = profiles_pb2.ResourceUsage()
        reader = graphsignal._agent.host_reader
        reader.read(resource_usage)
        HostReader.MIN_CPU_READ_INTERVAL = 0
        time.sleep(0.2)
        reader.read(resource_usage)

        #pp = pprint.PrettyPrinter()
        # pp.pprint(MessageToJson(resource_usage))

        self.assertTrue(resource_usage.host_usage.cpu_usage_percent > 0)
        self.assertTrue(resource_usage.host_usage.max_rss > 0)
        self.assertTrue(resource_usage.host_usage.current_rss > 0)
        self.assertTrue(resource_usage.host_usage.vm_size > 0)



import logging
import time
from threading import Lock

logger = logging.getLogger('graphsignal')

# global profiling lock
_profiling_lock = Lock()


class SpanScheduler(object):
    __slots__ = [
        '_total_span_count',
        '_ensured_span_count',
        '_span_filter',
        '_last_interval_ts'
    ]

    DEFAULT_SPANS = [10, 100, 1000]
    MAX_ENSURED_SPANS = 10
    MIN_SPAN_INTERVAL_SEC = 20

    def __init__(self):
        self._total_span_count = 0
        self._ensured_span_count = 0
        self._last_interval_ts = 0
        self._span_filter = {
            span: True for span in SpanScheduler.DEFAULT_SPANS}

    def lock(self, ensure=False):
        self._total_span_count += 1

        if _profiling_lock.locked():
            return False

        if ensure:
            self._ensured_span_count += 1
            if self._ensured_span_count > SpanScheduler.MAX_ENSURED_SPANS:
                return False
        else:
            # check if span index matches default span indexes
            if self._total_span_count not in self._span_filter:
                # skip first span
                if self._total_span_count == 1:
                    return False

                # comply with interval between spans
                if self._last_interval_ts > time.time() - \
                        self.MIN_SPAN_INTERVAL_SEC:
                    return False

        # set global lock
        return _profiling_lock.acquire(blocking=False)

    def unlock(self):
        if not _profiling_lock.locked():
            return
        self._last_interval_ts = time.time()
        _profiling_lock.release()



import unittest
import logging
import sys
import os
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.proto import profiles_pb2
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class PyTorchLightningTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            workload_name='w1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    @unittest.skip("enable manually")
    @patch.object(Uploader, 'upload_profile')
    def test_callback(self, mocked_upload_profile):
        import torch
        from pytorch_lightning import LightningModule, Trainer
        from torch import nn
        from torch.nn import functional as F
        from torch.utils.data import DataLoader, random_split
        from torchmetrics import Accuracy
        from torchvision import transforms
        from torchvision.datasets import MNIST
        from graphsignal.profilers.pytorch_lightning import GraphsignalProfiler

        PATH_DATASETS = os.environ.get("PATH_DATASETS", ".")
        AVAIL_GPUS = min(1, torch.cuda.device_count())
        BATCH_SIZE = 256 if AVAIL_GPUS else 64

        class MNISTModel(LightningModule):
            def __init__(self):
                super().__init__()
                self.l1 = torch.nn.Linear(28 * 28, 10)

            def forward(self, x):
                return torch.relu(self.l1(x.view(x.size(0), -1)))

            def training_step(self, batch, batch_nb):
                x, y = batch
                loss = F.cross_entropy(self(x), y)
                return loss

            def configure_optimizers(self):
                return torch.optim.Adam(self.parameters(), lr=0.02)

        # Init our model
        mnist_model = MNISTModel()

        # Init DataLoader from MNIST Dataset
        train_ds = MNIST(PATH_DATASETS, train=True, download=True, transform=transforms.ToTensor())
        train_loader = DataLoader(train_ds, batch_size=BATCH_SIZE)

        # Initialize a trainer
        trainer = Trainer(
            gpus=AVAIL_GPUS,
            max_epochs=3,
            profiler=GraphsignalProfiler()
        )

        # Train the model ⚡
        trainer.fit(mnist_model, train_loader)

        profile = mocked_upload_profile.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(profile))

        test_op_stats = None
        for op_stats in profile.op_stats:
            if op_stats.op_name == 'aten::addmm':
                test_op_stats = op_stats
                break
        self.assertIsNotNone(test_op_stats)
        self.assertTrue(test_op_stats.count >= 1)
        import torch
        if torch.cuda.is_available():
            self.assertTrue(test_op_stats.total_device_time_us >= 1)
            self.assertTrue(test_op_stats.self_device_time_us >= 1)
        else:
            self.assertTrue(test_op_stats.total_host_time_us >= 1)
            self.assertTrue(test_op_stats.self_host_time_us >= 1)



import os
from setuptools import setup, find_packages

def read(fname):
  return open(os.path.join(os.path.dirname(__file__), fname)).read()

version = {}
with open('graphsignal/version.py') as fp:
    exec(fp.read(), version)

setup(
  name = 'graphsignal',
  version = version['__version__'],
  description = 'Graphsignal Profiler',
  long_description = read('README.md'),
  long_description_content_type = 'text/markdown',
  author = 'Graphsignal, Inc.',
  author_email = 'devops@graphsignal.com',
  url = 'https://graphsignal.com',
  license = 'Apache-2.0',
  keywords = [
    'machine learning',
    'deep learning',
    'data science',
    'MLOps',
    'machine learning profiler',
    'deep learning profiler',
    'machine learning performance',
    'training profiling',
    'inference profiling'
  ],
  classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: Web Environment',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Topic :: Software Development',
    'Topic :: System :: Monitoring',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Artificial Intelligence'
  ],
  python_requires='>=3.5',
  install_requires=[
    'protobuf>3.0'
  ],
  
  packages = find_packages(exclude=[
    '*.sh', '*_test.py'])
)



import unittest
import logging
import sys
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.proto import profiles_pb2
from graphsignal.system_info import compare_semver

logger = logging.getLogger('graphsignal')


class SystemInfoTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            workload_name='w1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_compare_semver(self):
        s1 = profiles_pb2.SemVer()

        s1.major = 1
        s1.minor = 2
        self.assertEqual(compare_semver(s1, (1, 3, 0)), -1)

        s1.major = 1
        s1.minor = 2
        s1.patch = 3
        self.assertEqual(compare_semver(s1, (1, 2, 3)), 0)

        s1.major = 1
        s1.minor = 2
        s1.patch = 3
        self.assertEqual(compare_semver(s1, (1, 2, 2)), 1)



import time
import sys
import os
import logging
import threading
import uuid
import hashlib
import atexit

from graphsignal import version
from graphsignal.agent import Agent
from graphsignal.uploader import Uploader
from graphsignal.span_scheduler import SpanScheduler
from graphsignal.profiling_span import ProfilingSpan
from graphsignal.usage.host_reader import HostReader
from graphsignal.usage.nvml_reader import NvmlReader

logger = logging.getLogger('graphsignal')

_agent = None


def _check_configured():
    global _agent
    if not _agent:
        raise ValueError(
            'Graphsignal profiler not configured, call graphsignal.configure() first')


def configure(api_key, workload_name, debug_mode=False):
    global _agent

    if _agent:
        logger.warning('Graphsignal profiler already configured')
        return

    if debug_mode:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.WARNING)

    if not api_key:
        raise ValueError('Missing argument: api_key')

    if not workload_name:
        raise ValueError('Missing argument: workload_name')

    _agent = Agent()
    _agent.api_key = api_key
    _agent.run_id = _uuid_sha1(size=12)
    _agent.run_start_ms = int(time.time() * 1e3)
    _agent.workload_name = workload_name
    _agent.debug_mode = debug_mode
    _agent.uploader = Uploader()
    _agent.uploader.configure()
    _agent.host_reader = HostReader()
    _agent.host_reader.setup()
    _agent.nvml_reader = NvmlReader()
    _agent.nvml_reader.setup()
    _agent.span_scheduler = SpanScheduler()

    atexit.register(shutdown)

    logger.debug('Graphsignal profiler configured')


def shutdown():
    _check_configured()

    global _agent
    atexit.unregister(shutdown)
    _agent.uploader.flush()
    _agent.host_reader.shutdown()
    _agent.nvml_reader.shutdown()
    _agent = None

    logger.debug('Graphsignal profiler shutdown')


def _sha1(text, size=-1):
    sha1_hash = hashlib.sha1()
    sha1_hash.update(text.encode('utf-8'))
    return sha1_hash.hexdigest()[0:size]


def _uuid_sha1(size=-1):
    return _sha1(str(uuid.uuid4()), size)



import logging

from tensorflow.keras.callbacks import Callback

import graphsignal.profilers.tensorflow import profile_span

logger = logging.getLogger('graphsignal')


class GraphsignalCallback(Callback):
    __slots__ = [
        '_span'
    ]

    def __init__(self):
        self._span = None

    def _start_profiler(self, span_name, batch=None):
        if not self._span:
            self._span = profile_span(span_name=span_name)
            if batch is not None:
                self._span.add_metadata('batch', batch)

    def _stop_profiler(self):
        if self._span:
            self._span.stop()
            self._span = None

    def on_test_begin(self, logs=None):
        self._start_profiler('Test')

    def on_test_end(self, logs=None):
        self._stop_profiler()

    def on_predict_begin(self, logs=None):
        self._start_profiler('Prediction')

    def on_predict_end(self, logs=None):
        self._stop_profiler()

    def on_train_batch_begin(self, batch, logs=None):
        self._start_profiler('Training batch', batch=batch)

    def on_train_batch_end(self, batch, logs=None):
        self._stop_profiler()

    def on_test_batch_begin(self, batch, logs=None):
        self._start_profiler('Test batch', batch=batch)

    def on_test_batch_end(self, batch, logs=None):
        self._stop_profiler()

    def on_predict_batch_begin(self, batch, logs=None):
        self._start_profiler('Prediction batch', batch=batch)

    def on_predict_batch_end(self, batch, logs=None):
        self._stop_profiler()



import unittest
import logging
import sys
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.span_scheduler import SpanScheduler

logger = logging.getLogger('graphsignal')


class SpanSchedulerTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            workload_name='w1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_automatic(self):
        scheduler = SpanScheduler()
        self.assertFalse(scheduler.lock())
        scheduler.unlock()
        self.assertTrue(scheduler.lock())
        scheduler.unlock()
        self.assertFalse(scheduler.lock())
        scheduler.unlock()

    def test_default(self):
        scheduler = SpanScheduler()
        for _ in range(scheduler.DEFAULT_SPANS[0] - 1):
            scheduler.lock()
            scheduler.unlock()
        self.assertTrue(scheduler.lock())
        scheduler.unlock()

    def test_ensured(self):
        scheduler = SpanScheduler()
        self.assertTrue(scheduler.lock(ensure=True))
        scheduler.unlock()
        self.assertFalse(scheduler.lock(ensure=False))
        scheduler.unlock()
        self.assertTrue(scheduler.lock(ensure=True))
        scheduler.unlock()

    def test_concurrent(self):
        scheduler = SpanScheduler()
        self.assertTrue(scheduler.lock(ensure=True))
        self.assertFalse(scheduler.lock(ensure=True))
        self.assertFalse(scheduler.lock(ensure=True))
        scheduler.unlock()
        self.assertTrue(scheduler.lock(ensure=True))
        scheduler.unlock()



import logging
import os
import sys
import time
import tempfile
import gzip
import shutil
import glob
import tensorflow as tf
from tensorflow.python.client import device_lib
from google.protobuf.json_format import Parse

from graphsignal.profilers.tensorflow_proto import overview_page_pb2
from graphsignal.profilers.tensorflow_proto import input_pipeline_pb2
from graphsignal.profilers.tensorflow_proto import tf_stats_pb2
from graphsignal.profilers.tensorflow_proto import kernel_stats_pb2
from graphsignal.profilers.tensorflow_proto import memory_profile_pb2

import graphsignal
from graphsignal.system_info import parse_semver, compare_semver
from graphsignal.proto import profiles_pb2
from graphsignal.profiling_span import ProfilingSpan

logger = logging.getLogger('graphsignal')


class TensorflowProfiler():
    __slots__ = [
        '_is_initialized',
        '_log_dir',
        '_run_env'
    ]

    def __init__(self):
        self._is_initialized = False
        self._log_dir = None
        self._run_env = None

    def start(self):
        logger.debug('Activating TensorFlow profiler')

        if not self._is_initialized:
            logger.debug('Warming up TensorFlow profiler before first use')
            self._is_initialized = True
            tf.profiler.experimental.start('')
            tf.profiler.experimental.stop(save=False)
            logger.debug('Finished warming up')

            self._read_run_env()

        try:
            self._create_log_dir()

            options = tf.profiler.experimental.ProfilerOptions(
                host_tracer_level=2,
                python_tracer_level=1,
                device_tracer_level=1)
            tf.profiler.experimental.start(self._log_dir, options=options)
        except Exception as e:
            self._remove_log_dir()
            raise e

        return True

    def stop(self, profile):
        logger.debug('Deactivating TensorFlow profiler')

        try:
            tf.profiler.experimental.stop()

            self._copy_run_env(profile)
            self._convert_to_profile(profile)
        except Exception as e:
            raise e
        finally:
            self._remove_log_dir()

        return True

    def _create_log_dir(self):
        self._log_dir = tempfile.mkdtemp(prefix='graphsignal-')
        logger.debug('Created temporary log directory %s', self._log_dir)

    def _remove_log_dir(self):
        shutil.rmtree(self._log_dir)
        logger.debug('Removed temporary log directory %s', self._log_dir)

    def _read_run_env(self):
        self._run_env = profiles_pb2.RunEnvironment()
        self._run_env.ml_framework = profiles_pb2.RunEnvironment.MLFramework.TENSORFLOW
        parse_semver(self._run_env.ml_framework_version, tf.__version__)
        if compare_semver(self._run_env.ml_framework_version, (2, 2, 0)) == -1:
            raise Exception(
                'TensorFlow profiling is not supported for versions <=2.2')
        for device in tf.config.list_physical_devices('GPU'):
            device_proto = self._run_env.devices.add()
            device_proto.type = profiles_pb2.DeviceType.GPU
            details = tf.config.experimental.get_device_details(device)
            if details:
                if 'device_name' in details:
                    device_proto.name = details['device_name']
                if 'compute_capability' in details:
                    device_proto.is_cuda_enabled = True
                    device_proto.compute_capability.major = details['compute_capability'][0]
                    device_proto.compute_capability.minor = details['compute_capability'][1]

    def _copy_run_env(self, profile):
        profile.run_env.ml_framework = self._run_env.ml_framework
        profile.run_env.ml_framework_version.CopyFrom(
            self._run_env.ml_framework_version)
        profile.run_env.devices.extend(self._run_env.devices)

    def _convert_to_profile(self, profile):
        overview_page_data = self._find_and_read(
            'plugins/profile/*/*overview_page.pb')
        if overview_page_data:
            overview_page = overview_page_pb2.OverviewPage()
            overview_page.ParseFromString(overview_page_data)
            analysis = overview_page.analysis
            profile.summary.device_compute_16bit_percent = analysis.device_compute_16bit_percent
            profile.summary.device_compute_32bit_percent = analysis.device_compute_32bit_percent
            profile.summary.mxu_utilization = analysis.mxu_utilization_percent

            input_analysis = overview_page.input_analysis
            profile.summary.input_percent = input_analysis.input_percent
            profile.summary.output_percent = input_analysis.output_percent

        host_idle_time_us = 0
        device_idle_time_us = 0
        tf_stats_data = self._find_and_read(
            'plugins/profile/*/*tensorflow_stats.pb')
        if tf_stats_data:
            tf_stats_db = tf_stats_pb2.TfStatsDatabase()
            tf_stats_db.ParseFromString(tf_stats_data)
            for tf_stats_record in tf_stats_db.with_idle.tf_stats_record:
                if tf_stats_record.op_type == 'IDLE':
                    if tf_stats_record.host_or_device == 'Host':
                        host_idle_time_us += int(tf_stats_record.total_self_time_in_us)
                    else:
                        device_idle_time_us += int(tf_stats_record.total_self_time_in_us)
                    continue
                op_stats = profile.op_stats.add()
                if tf_stats_record.host_or_device == 'Host':
                    op_stats.device_type = profiles_pb2.DeviceType.CPU
                    op_stats.total_host_time_us = int(tf_stats_record.total_time_in_us)
                    op_stats.self_host_time_us = int(tf_stats_record.total_self_time_in_us)
                    op_stats.self_host_memory_rate = int(tf_stats_record.measured_memory_bw)
                else:
                    op_stats.device_type = profiles_pb2.DeviceType.GPU
                    op_stats.total_device_time_us = int(tf_stats_record.total_time_in_us)
                    op_stats.self_device_time_us = int(tf_stats_record.total_self_time_in_us)
                    op_stats.self_device_memory_rate = int(tf_stats_record.measured_memory_bw)
                    op_stats.tensorcore_utilization = tf_stats_record.gpu_tensorcore_utilization
                op_stats.op_type = tf_stats_record.op_type
                op_stats.op_name = tf_stats_record.op_name
                op_stats.count = int(tf_stats_record.occurrences)
        else:
            logger.debug('No operation data found in TensorFlow log directory')

        kernel_stats_data = self._find_and_read(
            'plugins/profile/*/*kernel_stats.pb')
        if kernel_stats_data:
            kernel_stats_db = kernel_stats_pb2.KernelStatsDb()
            kernel_stats_db.ParseFromString(kernel_stats_data)
            for kernel_report in kernel_stats_db.reports:
                kernel_stats = profile.kernel_stats.add()
                kernel_stats.device_type = profiles_pb2.DeviceType.GPU
                kernel_stats.op_name = kernel_report.op_name
                kernel_stats.kernel_name = kernel_report.name
                kernel_stats.count = int(kernel_report.occurrences)
                kernel_stats.duration_ns = int(kernel_report.total_duration_ns)
                kernel_stats.is_using_tensorcore = kernel_report.is_kernel_using_tensor_core
        else:
            logger.debug('No kernel data found in TensorFlow log directory')

        sum_host_op_time_us = 0
        sum_device_op_time_us = 0
        for op_stats in profile.op_stats:
            sum_host_op_time_us += op_stats.self_host_time_us
            sum_device_op_time_us += op_stats.self_device_time_us
        sum_op_time_us = sum_host_op_time_us + sum_device_op_time_us
        if sum_op_time_us > 0:
            profile.summary.host_op_percent = sum_host_op_time_us / sum_op_time_us * 100
            profile.summary.device_op_percent = sum_device_op_time_us / sum_op_time_us * 100

        if host_idle_time_us > 0:
            sum_host_op_time_us_with_idle = sum_host_op_time_us + host_idle_time_us
            if sum_host_op_time_us_with_idle > 0:
                profile.summary.host_idle_percent = host_idle_time_us / sum_host_op_time_us_with_idle * 100
        if device_idle_time_us > 0:
            sum_device_op_time_us_with_idle = sum_device_op_time_us + device_idle_time_us
            if sum_device_op_time_us_with_idle > 0:
                profile.summary.device_idle_percent = device_idle_time_us / sum_device_op_time_us_with_idle * 100


    def _find_and_read(self, file_pattern):
        file_paths = glob.glob(os.path.join(self._log_dir, file_pattern))
        if len(file_paths) == 0:
            raise Exception(
                'Files are not found at {}'.format(
                    os.path.join(
                        self._log_dir,
                        file_pattern)))

        if file_paths[-1].endswith('.gz'):
            last_file = gzip.open(file_paths[-1], "rb")
        else:
            last_file = open(file_paths[-1], "rb")
        data = last_file.read()
        last_file.close()

        return data


_profiler = TensorflowProfiler()

def profile_span(span_name=None, ensure_profile=False):
    graphsignal._check_configured()

    return ProfilingSpan(
        scheduler=graphsignal._agent.span_scheduler,
        profiler=_profiler,
        span_name=span_name,
        ensure_profile=ensure_profile)
